/**==========================================================================

	PC-8031 Device Emulator for Win32 Ver 0.90
	Copyright (C)1993-2024 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	last modified  Apr,27,2024.

;==========================================================================*/
#include "../PC-8001/PC-8001.h"
#ifdef __cplusplus
extern "C"
{
#endif
extern void __cdecl	__z80_set_external_method(LPZ80CPU cpu, LPZ80FNCMBR mbr, LPZ80FNCMBW mbw, LPZ80FNCIOR ior, LPZ80FNCIOW iow); 
extern void _fdu_init_ioport(void);					// FDU_IO.c
extern void _fdu_init_memory(void);					// FDU_MEMORY.c
extern void _fdu_init_fduif(void);					// PD8255.c
extern void _fdu_init_fdc(LPFDU765A, LPFDU);		// PD765A.c
extern void _fdu_init_drives(LPFDUDRIVES, LPFDU);	// FDD.c
extern BOOL _fdu_image_scan(LPFDU, LPDISKIMAGE);	// MEDIA.c
#define export
#define SUBCPULIFE	(15000)							// 15sec(unit:msec)
/**--------------------------------------------------------------------------
 **		Local Variables
 **------------------------------------------------------------------------*/
static int (*save_step)(LPZ80CPU cpu) = NULL;
/**--------------------------------------------------------------------------
 **		Proxy Functions
 **------------------------------------------------------------------------*/
static BYTE __cdecl _fdu_mbread(LPZ80CPU cpu, USHORT Adr, DWORD* W) {
	return n80device.fdu.mem.read(&n80device.fdu.mem, Adr, W);
}
static void __cdecl _fdu_mbwrite(LPZ80CPU cpu, USHORT Adr, BYTE v, DWORD* W) {
	n80device.fdu.mem.write(&n80device.fdu.mem, Adr, v, W);
}
static BYTE __cdecl _fdu_ioread(LPZ80CPU cpu, USHORT port) {
	return n80device.fdu.io.read(&n80device.fdu, port);
}
static void __cdecl _fdu_iowrite(LPZ80CPU cpu, USHORT port, BYTE v) {
	n80device.fdu.io.write(&n80device.fdu, port, v);
}
/**--------------------------------------------------------------------------
 **		Z80 override Methods
 **------------------------------------------------------------------------*/
static int __cdecl __z80_step_override_nop(LPZ80CPU cpu) {
	return 0;
}
static int __cdecl __z80_step_override_active(LPZ80CPU cpu) {
	LPFDU fdu = &n80device.fdu;
	int result = 0;
	DWORD Ts = 0;

	// sync between main and sub cpu state count
	while (fdu->cpu_sync_T > 0) {
		result = save_step(cpu);
		fdu->cpu.r.x.T += fdu->cpu.r.x.TwM1;
		fdu->cpu_sync_T -= (long)fdu->cpu.r.x.T;
		Ts += fdu->cpu.r.x.T;
		// if sub cpu running at RAM area (not at ROM) then extend life-cycle
		if (fdu->cpu.r.x.PC >= 0x4000)
			fdu->cpu_T_limit = fdu->msec_to_T(fdu, SUBCPULIFE);	// 15sec
		// feed the current session a exe/result-phase run-time clock
		if (fdu->fdc.current != NULL &&
			fdu->fdc.current->accept_clk && (
			fdu->fdc.current->hold_int ||
			fdu->fdc.phase & (FDC_PHASE_EXE | FDC_PHASE_RESULT))) {
			DWORD T = fdu->cpu_sleeping ? 4 : cpu->r.x.T;
			fdu->resume(fdu);
			fdu->fdc.current->clk(fdu->fdc.current, T, FDC_CLK_BY_CLK);
		}
	}
	if (fdu->cpu_T_limit > Ts)
		fdu->cpu_T_limit -= Ts;
	else
		fdu->cpu_T_limit = 0;

	// enter sub cpu sleep
	if (fdu->cpu_T_limit == 0) {
		// IDOL state and no session (except for fdu RAM routine is executing)
		if (fdu->fdc.phase == FDC_PHASE_IDOL && fdu->fdc.current == NULL && fdu->cpu.r.x.PC < 0x4000) {
			fdu->cpu_sleeping = TRUE;
			fdu->cpu.step = __z80_step_override_nop;
#if _DEBUG
			n80device.logger("FDU: [CPU] enter cpu sleep mode.\n");
#endif
		} else	// auto-extend cpu time limit
			fdu->cpu_T_limit = fdu->msec_to_T(fdu, SUBCPULIFE);	// 15sec
	}
	return result;
}
/**--------------------------------------------------------------------------
 **		\bh`
 **------------------------------------------------------------------------*/
static void _fdu_set_rom_ready(LPFDU self, BOOL flag) {
	self->rom_ready = flag;
	self->cpu_T_limit = self->msec_to_T(self, SUBCPULIFE);	// 15sec
    if (save_step == NULL)
		save_step = self->cpu.step;
	// if DISK ROM is not ready then let FDU Z80 be always disable
	self->cpu.step = flag ? __z80_step_override_active : __z80_step_override_nop;
}
static void _fdu_reset(LPFDU self) {
	self->fdc.reset(&self->fdc);
	self->fdd.reset(&self->fdd);
	self->mem.clear(&self->mem);
	self->io.clear(self);
	self->cpu.reset(&self->cpu);
	self->cpu_sync_T = 0;
}
static BOOL _fdu_available(LPFDU self) {
	return self->rom_ready &&
		(self->fdd.drives[0].media_inserted || self->fdd.drives[1].media_inserted);
}
static void _fdu_resume(LPFDU self) {
	if (self->available(self)) {
		self->cpu_T_limit = self->msec_to_T(self, SUBCPULIFE);	// 15sec
		if (self->cpu_sleeping) {
			self->cpu_sleeping = FALSE;
			self->cpu.step = __z80_step_override_active;
			self->cpu_sync_T = 0;
#if _DEBUG
			n80device.logger("FDU: [CPU] wake up cpu.\n");
#endif
		}
	}
}
static DWORD _fdu_msec_to_T(LPFDU self, double msec) {
	return self->owner->msec_to_T(self->owner, msec);
}
static void _fdu_dispose(LPFDU self) {
	for (int i = 0; i < self->fdd.drives_cnt; ++i) {
		self->fdd.drives[i].eject(&self->fdd.drives[i]);
	}
}
/**--------------------------------------------------------------------------
 **		Initialize
 **------------------------------------------------------------------------*/
export void __cdecl _fdu_initialize(LPFDU self, LPN80DEVICE owner) {
	self->owner				= owner;
	self->dispose			= _fdu_dispose;
	self->mem.init			= _fdu_init_memory;
	self->io.init			= _fdu_init_ioport;
	self->ppi.init			= _fdu_init_fduif;
	self->fdc.init			= _fdu_init_fdc;
	self->fdd.init			= _fdu_init_drives;
	self->reset				= _fdu_reset;
	self->set_rom_ready		= _fdu_set_rom_ready;
	self->available			= _fdu_available;
	self->resume			= _fdu_resume;
	self->msec_to_T			= _fdu_msec_to_T;
	self->media_init		= _fdu_image_scan;
	self->rom_ready			= FALSE;
	self->cpu_sleeping		= FALSE;

	self->mem.init();
	self->io.init();
	self->ppi.init();
	self->fdc.init(&self->fdc, self);
	self->fdd.init(&self->fdd, self);

	__z80_set_external_method(&self->cpu, _fdu_mbread, _fdu_mbwrite, _fdu_ioread, _fdu_iowrite);
	self->cpu.id = 2;	// main = 1, sub = 2
	self->reset(self);
}

#ifdef __cplusplus
}	// extern "C"
#endif
