/**==========================================================================

	PC-8031 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2024 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	last modified  Apr,27,2024.

;==========================================================================*/
#include "../PC-8001/PC-8001.h"
#ifdef __cplusplus
extern "C"
{
#endif
#define export
/**--------------------------------------------------------------------------
 **		Base Methods
 **------------------------------------------------------------------------*/
static void _ses_empty_command(LPSESSION self, BYTE data) {
#if _DEBUG
	n80device.logger("FDU: [CMD] \"%s COMMAND\" has NO MORE COMMAND PHASE !!! \n", self->name);
#endif
}
static BYTE _ses_empty_read_data(LPSESSION self) {
#if _DEBUG
	n80device.logger("FDU: [EXE] \"%s COMMAND\" has NO READ/RESULT PHASE !!! \n", self->name);
#endif
	return 0xff;
}
static BYTE _ses_read_result_data(LPSESSION self) {
	BYTE data = 0xff;
	if (self->fdc->phase == FDC_PHASE_RESULT) {
		if (self->fdc->result.count > 0) {
#if _DEBUG
			if (self->fdc->result.index == 0) {
				for (int i = 0; i < self->fdc->result.count; ++i) {
					n80device.logger("<%2.2X>", self->fdc->result.ptr[i]);
				}
				n80device.logger("\n");
			}
#endif
			data = self->fdc->result.pop(&self->fdc->result);
#if _DEBUG
			//n80device.logger("FDU: [RES] 0x%2.2X (%d)\n", data, data);
#endif
		}
		if (self->fdc->result.count == 0) {
			// 2) enter == IDOL PHASE ==
			self->fdc->enter_idol_phase(self->fdc);
			if (self->fdc->owner->cpu.signal & _HLTBIT && self->fdc->owner->cpu.r.x.IFF1 == 1)
				self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_);
		}
	}
	return data;
}

static void _ses_empty_write_data(LPSESSION self, BYTE byte) {
#if _DEBUG
	n80device.logger("FDU: [EXE] \"%s COMMAND\" has NO WRITE PHASE !!! \n", self->name);
#endif
}
static void _ses_empty_clk(LPSESSION self, DWORD T, BYTE tc) {
	//if (self->fdc->owner->cpu.halt_detect)
	//	self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_);
}
static BOOL _ses_raise_int(LPSESSION self) {
	if (self->fdc->owner->cpu.signal & _HLTBIT) {
		if (self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_) == 0) {
			return TRUE;
		}
	}
	return FALSE;
}
static void _ses_start_common_session(LPSESSION self, LPFDU765A fdc) {
	self->fdc			= fdc;
	self->accept_clk	= TRUE;
	self->hold_int		= FALSE;
	self->command		= _ses_empty_command;
	self->read_data		= _ses_empty_read_data;
	self->write_data	= _ses_empty_write_data;
	self->clk			= _ses_empty_clk;
	self->raise_int		= _ses_raise_int;
	fdc->current		= self;
}
#ifdef _MSC_VER
#pragma region SPECIFY COMMAND SESSION
#else
// #region SPECIFY COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					SPECIFY 
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		--------	---------------------------------------------------------
 **		COMMAND		[0x03], [SRT/HUT], [HLT/ND]
 **		EXECUTION	--
 **		RESULT		--
 **------------------------------------------------------------------------*/
static void _ses_specify_command(LPSESSION self, BYTE data) {
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (cmd->push(cmd, data) == 3) {
		cmd->load_spec(cmd);
		self->fdc->step_rate_time	= (int)cmd->SRT + 1;		//(msec)
		self->fdc->head_unload_time = ((int)cmd->HUT + 1) * 16;	//(msec)
		self->fdc->head_load_time	= ((int)cmd->HLT + 1) * 16;	//(msec)
		self->fdc->non_dma_mode		= cmd->ND == 1;
#if _DEBUG
		n80device.logger("FDU: [CMD] %s %2.2X, %2.2X [SRT:%d, HUT:%d, HLT:%d, NonDMA:%s]\n",
			self->name, cmd->commands[1], cmd->commands[2],
			self->fdc->step_rate_time, self->fdc->head_unload_time, self->fdc->head_load_time,
			self->fdc->non_dma_mode ? "Y" : "N");
#endif
		// 2) enter == IDOL PHASE ==
		self->fdc->enter_idol_phase(self->fdc);
	}
}

export void _ses_start_specify_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "SPECIFY";
	self->fdc			= fdc;
	self->accept_clk	= FALSE;
	self->command		= _ses_specify_command;
	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region SEEK COMMAND SESSION
#else
// #region SEEK COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					SEEK
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x0F], [HD/US0/US1], [NCN]
 **		EXECUTION	head retracked to new cylinder
 **		RESULT		--
 **------------------------------------------------------------------------*/
static void _ses_seek_command(LPSESSION self, BYTE data) {
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (cmd->push(cmd, data) == 3) {
		cmd->load_seek(cmd);
		LPFDUDRIVES fdd = &self->fdc->owner->fdd;
		LPFDUDRIVE  drv = &fdd->drives[cmd->US];
		drv->interval = fdd->seek(fdd, cmd->US, cmd->NCN, cmd->HD);
#if _DEBUG
		n80device.logger("FDU: [CMD] %s %2.2X %2.2X [UNIT:%d, HEAD:%d, NCN:%d]\n",
			self->name, cmd->commands[1], cmd->commands[2], cmd->US, cmd->HD, cmd->NCN);
#endif
		// 2) enter == EXECUTION PHASE ==
		self->fdc->enter_exe_phase(self->fdc, cmd->US);
		self->fdc->status &= ~FDC_ST_CB;				// SEEK command accept another drive SEEK
		self->fdc->status |= (BYTE)(0x01 << cmd->US);	// DB0..4 (FDD n Busy in SEEK mode)
	}
}

static void _ses_seek_clk(LPSESSION self, DWORD T, BYTE tc) {
	LPFDUDRIVES fdd = &self->fdc->owner->fdd;
	LPFDUDRIVE	drv = &fdd->drives[self->fdc->cmd.US];
	if (drv->interval == 0) {
		if (self->fdc->phase != FDC_PHASE_EXE)
			return;
		// break cpu HALT
		if (self->fdc->owner->cpu.signal & _HLTBIT) {
			if (self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_) != 0) {
				drv->interval = 100;
				return;
			}
		}
		// 3) enter == IDOL PHASE ==
		self->fdc->enter_idol_phase(self->fdc);
		self->fdc->status &= ~FDC_ST_EXM;	// clear EXM
	}
	else if (drv->interval > T)
		drv->interval -= T;
	else
		drv->interval = 0;
}

export void _ses_start_seek_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "SEEK";
	self->command		= _ses_seek_command;
	self->clk			= _ses_seek_clk;
	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
	// #endregion
#endif
#ifdef _MSC_VER
#pragma region RECALIBRATE COMMAND SESSION
#else
// #region RECALIBRATE COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					RECALIBRATE
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x07], [US0/US1]
 **		EXECUTION	head retracked to track 0
 **		RESULT		--
 **------------------------------------------------------------------------*/
static void _ses_recalibrate_command(LPSESSION self, BYTE data) {
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (cmd->push(cmd, data) == 2) {
		cmd->load_cmn(cmd);
		LPFDUDRIVES fdd = &self->fdc->owner->fdd;
		LPFDUDRIVE  drv = &fdd->drives[cmd->US];
		drv->interval = fdd->seek(fdd, cmd->US, 0, 0);
#if _DEBUG
		n80device.logger("FDU: [CMD] %s %2.2X [UNIT:%d]\n", self->name, cmd->commands[1], cmd->US);
#endif
		// 2) enter == EXECUTION PHASE ==
		self->fdc->enter_exe_phase(self->fdc, cmd->US);
	}
}

export void _ses_start_recalibrate_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "RECALIBRATE";
	self->command		= _ses_recalibrate_command;
	self->clk			= _ses_seek_clk;
	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region SENSE INTERRUPT STATUS COMMAND SESSION
#else
// #region SENSE INTERRUPT STATUS COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					SENSE INTERRUPT STATUS
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x08]
 **		EXECUTION	--
 **		RESULT		[ST0], [PCN]
 **------------------------------------------------------------------------*/
export void _ses_start_sns_int_sts_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "SENSE INTERRUPT STATUS";
	self->accept_clk	= FALSE;
	self->read_data		= _ses_read_result_data;

	self->fdc->owner->fdd.snsint(&self->fdc->owner->fdd);
#if _DEBUG
	n80device.logger("FDU: [CMD] %s\n", self->name);
#endif
	// 1) enter == RESULT PHASE ==
	self->fdc->enter_result_phase(fdc);
	fdc->status &= 0xf0;	// clear SEEKing or holding SEEK complete INT bit
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region SENSE DRIVE STATUS COMMAND SESSION
#else
// #region SENSE DRIVE STATUS COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					SENSE DRIVE STATUS
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x04], [HD/US0/US1]
 **		EXECUTION	--
 **		RESULT		[ST3]
 **------------------------------------------------------------------------*/
static void _ses_sns_drv_sts_command(LPSESSION self, BYTE data) {
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (cmd->push(cmd, data) == 2) {
		cmd->load_cmn(cmd);
		LPFDUDRIVES fdd = &self->fdc->owner->fdd;
		fdd->snsdrv(fdd, cmd->US, cmd->HD);
#if _DEBUG
		n80device.logger("FDU: [CMD] %s %2.2X [UNIT:%d, HD:%d]\n", self->name, cmd->commands[1], cmd->US, cmd->HD);
#endif
		// 2) enter == RESULT PHASE ==
		self->fdc->enter_result_phase(self->fdc);
	}
}

export void _ses_start_sns_drv_sts_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "SENSE DRIVE STATUS";
	self->accept_clk	= FALSE;
	self->command		= _ses_sns_drv_sts_command;
	self->read_data		= _ses_read_result_data;

	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region READ DATA COMMAND SESSION
#else
// #region READ DATA COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					READ DATA
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x06/MT/MF/SK], [HD/US0/US1], [C], [H], [R], [N], [EOT], [GPL], [DTL]
 **		EXECUTION	data read from disk
 **		RESULT		[ST0], [ST1], [ST2], [C], [H], [R], [N]
 **------------------------------------------------------------------------*/
static void _ses_read_data_command(LPSESSION self, BYTE data) {
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (cmd->push(cmd, data) == 9) {
		cmd->load_std(cmd);
		LPFDUDRIVES fdd = &self->fdc->owner->fdd;
		LPFDUDRIVE  drv = &fdd->drives[cmd->US];
		self->fdc->rw_data_cnt = 0;
		self->multi_trk_count = 0;
		drv->remain_cnt = 0;
		drv->data_buff_index = 0;
		self->fdc->result.let_std(&self->fdc->result, cmd);
#if _DEBUG
		n80device.logger("FDU: [CMD] %s [UNIT:%d, C:%d(%d), H:%d, R:%d, N:%d, EOT:%d, GPL:%d, DTL:%d, HD:%d, MT:%d]\n",
			self->name, cmd->US, drv->cylinder, cmd->id.C, cmd->id.H, cmd->id.R, cmd->id.N, cmd->EOT, cmd->GPL, cmd->DTL, cmd->HD, cmd->MT);
#endif
		// 2) enter == EXECUTION PHASE ==
		self->fdc->enter_exe_phase(self->fdc, cmd->US);
	}
}
static void _ses_read_data_clk(LPSESSION self, DWORD T, BYTE tc) {
	LPFDUDRIVES fdd = &self->fdc->owner->fdd;
	LPFDUDRIVE drv = &fdd->drives[self->fdc->cmd.US];
	// call by TC
	if (tc == FDC_CLK_BY_TC) {
		if (self->fdc->phase == FDC_PHASE_EXE) {
#if _DEBUG
			n80device.logger("     {READ DATA COUNT} %d <TC>\n", self->data_count);
#endif
			// 3) enter == RESULT PHASE ==
			self->fdc->result.push_std(&self->fdc->result);
			self->fdc->enter_result_phase(self->fdc);
			drv->interval = 50;
			return;
		} else if (self->fdc->phase == FDC_PHASE_RESULT) {
			if (self->multi_trk_count > 0 && self->fdc->result.ST0 & 0xc0) {
				self->fdc->result.restore(&self->fdc->result, &self->fdc->last_result);
			}
			drv->interval = 50;
			return;
		} else {
			// 4) enter == IDOL PHASE ==
			self->fdc->enter_idol_phase(self->fdc);
		}
		// release CPU HALT state with INT signal
		self->raise_int(self);
		return;
	}
	// call by clk
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (drv->interval == 0) {
		drv->interval = 50;
		if (self->fdc->phase == FDC_PHASE_EXE) {
			if (drv->remain_cnt == 0) {
				if (fdd->read(fdd, self->fdc->cmd.US, cmd, self->multi_trk_count) < 0) {
#if _DEBUG
					n80device.logger("     {READ DATA COUNT %d}\n", self->data_count);
#endif
					// 3) enter == RESULT PHASE ==
					self->fdc->result.push_std(&self->fdc->result);
					self->fdc->enter_result_phase(self->fdc);
					return;
				}
				self->multi_trk_count++;
				self->fdc->result.save(&self->fdc->result, &self->fdc->last_result);
			}
			self->fdc->status |= (FDC_ST_RQM | FDC_ST_DIO);
		}
		// release CPU HALT state with INT signal
		self->raise_int(self);
	}
	else if (drv->interval > T)
		drv->interval -= T;
	else
		drv->interval = 0;
}

static BYTE _ses_read_data(LPSESSION self) {
	LPFDUDRIVES fdd = &self->fdc->owner->fdd;
	LPFDUDRIVE	drv = &fdd->drives[self->fdc->cmd.US];
	BYTE data = 0xff;

	if (self->fdc->phase == FDC_PHASE_EXE) {
		if (drv->remain_cnt > 0) {
			drv->remain_cnt--;
			data = drv->data_buff[drv->data_buff_index++];
			++self->data_count;
			self->fdc->status &= ~FDC_ST_RQM;
			self->fdc->status &= ~FDC_ST_DIO;
#if _DEBUG
			n80device.logger("%2.2X,", data);
			if ((self->data_count % 16) == 0)
				n80device.logger("\n");
		} else {
			n80device.logger("     {### READ DATA NOT READY ###}]\n");
#endif
		}
	} else {
		if (self->fdc->phase == FDC_PHASE_RESULT) {
			if (self->fdc->result.count > 0) {
#if _DEBUG
				if (self->fdc->result.index == 0) {
					for (int i = 0; i < self->fdc->result.count; ++i) {
						n80device.logger("<%2.2X>", self->fdc->result.ptr[i]);
					}
					n80device.logger("\n");
				}
#endif
				data = self->fdc->result.pop(&self->fdc->result);
			}
			if (self->fdc->result.count == 0) {
				// 2) enter == IDOL PHASE ==
				self->fdc->enter_idol_phase(self->fdc);
				// release CPU HALT state with INT signal
				if (self->fdc->owner->cpu.signal & _HLTBIT && self->fdc->owner->cpu.r.x.IFF1 == 1)
					self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_);
			}
		}
	}
	return data;
}

export void _ses_start_read_data_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name				= "READ DATA";
	self->command			= _ses_read_data_command;
	self->read_data			= _ses_read_data;
	self->clk				= _ses_read_data_clk;
	self->multi_trk_count	= 0;
	self->data_count		= 0;

	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region READ DELETED DATA COMMAND SESSION
#else
// #region READ DELETED DATA COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					READ DELETED DATA
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x0C/MT/MF/SK], [HD/US0/US1], [C], [H], [R], [N], [EOT], [GPL], [DTL]
 **		EXECUTION	data deleted read from disk
 **		RESULT		[ST0], [ST1], [ST2], [C], [H], [R], [N]
 **------------------------------------------------------------------------*/
export void _ses_start_read_del_data_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name				= "READ DELETED DATA";
	self->command			= _ses_read_data_command;
	self->read_data			= _ses_read_data;
	self->clk				= _ses_read_data_clk;
	self->multi_trk_count	= 0;
	self->data_count		= 0;

	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region READ DIAGNOSTIC COMMAND SESSION
#else
// #region READ DIAGNOSTIC COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					READ DIAGNOSTIC
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x02/MF/SK], [HD/US0/US1], [C], [H], [R], [N], [EOT], [GPL], [DTL]
 **		EXECUTION	all data read from disk
 **		RESULT		[ST0], [ST1], [ST2], [C], [H], [R], [N]
 **------------------------------------------------------------------------*/
static void _ses_read_diag_clk(LPSESSION self, DWORD T, BYTE tc) {
	// call by TC
	if (tc == FDC_CLK_BY_TC) {
		if (self->fdc->phase == FDC_PHASE_EXE) {
#if _DEBUG
			n80device.logger("     {READ DIAG COUNT} %d\n", self->data_count);
#endif
			// 3) enter == RESULT PHASE ==
			self->fdc->result.push_std(&self->fdc->result);
			self->fdc->enter_result_phase(self->fdc);
		} else if (self->fdc->phase == FDC_PHASE_RESULT) {
			;
		} else {
			// 4) enter == IDOL PHASE ==
			self->fdc->enter_idol_phase(self->fdc);
		}
		// release CPU HALT state with INT signal
		if (self->fdc->owner->cpu.signal & _HLTBIT && self->fdc->owner->cpu.r.x.IFF1 == 1)
			self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_);
		return;
	}
	// call by clk
	LPFDUDRIVES fdd = &self->fdc->owner->fdd;
	LPFDUDRIVE drv = &fdd->drives[self->fdc->cmd.US];
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (drv->interval == 0) {
		drv->interval = 100;
		if (self->fdc->phase == FDC_PHASE_EXE) {
			if (drv->remain_cnt == 0) {
				if (fdd->diagnostic(fdd, self->fdc->cmd.US, cmd, self->multi_trk_count++) < 0) {
#if _DEBUG
					n80device.logger("     {READ DIAG COUNT %d}\n", self->data_count);
#endif
					// 3) enter == RESULT PHASE ==
					self->fdc->result.push_std(&self->fdc->result);
					self->fdc->enter_result_phase(self->fdc);
				}
			}
			self->fdc->status |= (FDC_ST_RQM | FDC_ST_DIO);
		}
		// release CPU HALT state with INT signal
		if (self->fdc->owner->cpu.signal & _HLTBIT && self->fdc->owner->cpu.r.x.IFF1 == 1) {
			self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_);
		}
	}
	else if (drv->interval > T)
		drv->interval -= T;
	else
		drv->interval = 0;
}

export void _ses_start_read_diag_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name				= "READ DIAGNOSTIC";
	self->command			= _ses_read_data_command;
	self->read_data			= _ses_read_data;
	self->clk				= _ses_read_diag_clk;
	self->multi_trk_count	= 0;
	self->data_count		= 0;

	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region READ ID COMMAND SESSION
#else
// #region READ ID COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					READ ID
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		----------	---------------------------------------------------------
 **		COMMAND		[0x0A/MF], [HD/US0/US1]
 **		EXECUTION	--
 **		RESULT		[ST0], [ST1], [ST2], [C], [H], [R], [N]
 **------------------------------------------------------------------------*/
static void _ses_read_id_command(LPSESSION self, BYTE data) {
	LPFDUCOMMAND cmd = &self->fdc->cmd;
	if (cmd->push(cmd, data) == 2) {
		cmd->load_cmn(cmd);
		LPFDUDRIVES fdd = &self->fdc->owner->fdd;
		LPFDUDRIVE  drv = &fdd->drives[cmd->US];
		drv->interval = 120;
#if _DEBUG
		n80device.logger("FDU: [CMD] %s %2.2X [UNIT:%d, HD:%d]\n", self->name, cmd->commands[1], cmd->US, cmd->HD);
#endif
		// 2) enter == EXECUTION PHASE ==
		self->fdc->enter_exe_phase(self->fdc, cmd->US);
	}
}
static void _ses_read_id_clk(LPSESSION self, DWORD T, BYTE tc) {
	if (self->fdc->phase == FDC_PHASE_EXE) {
		LPFDUDRIVES fdd = &self->fdc->owner->fdd;
		LPFDUDRIVE drv = &fdd->drives[self->fdc->cmd.US];
		LPFDUCOMMAND cmd = &self->fdc->cmd;
		if (drv->interval == 0 || tc == FDC_CLK_BY_TC) {
			fdd->read_id(fdd, cmd->US, cmd);
			self->fdc->result.push_std(&self->fdc->result);
			// 3) enter == RESULT PHASE ==
			self->fdc->result.push_std(&self->fdc->result);
			self->fdc->enter_result_phase(self->fdc);
			self->fdc->status |= (FDC_ST_RQM | FDC_ST_DIO);
		}
		else if (drv->interval > T)
			drv->interval -= T;
		else
			drv->interval = 0;
	} else if (self->fdc->phase == FDC_PHASE_RESULT) {
		// release CPU HALT state with INT signal
		if (self->fdc->owner->cpu.signal & _HLTBIT && self->fdc->owner->cpu.r.x.IFF1 == 1) {
			self->fdc->owner->cpu.int_signal(&self->fdc->owner->cpu, _Op_NOP_);
		}
	}
}

export void _ses_start_read_id_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "READ ID";
	self->command		= _ses_read_id_command;
	self->read_data		= _ses_read_result_data;
	self->clk			= _ses_read_id_clk;

	// 1) enter == COMMAND PHASE ==
	fdc->enter_cmd_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
#ifdef _MSC_VER
#pragma region INVALID COMMAND SESSION
#else
// #region INVALID COMMAND SESSION
#endif
/**--------------------------------------------------------------------------
 **					INVALID
 **		---------------------------------------------------------------------
 **		PHASE		Instruction Code
 **		--------	---------------------------------------------------------
 **		RESULT		[0x80]
 **------------------------------------------------------------------------*/
export void _ses_start_invalid_session(LPSESSION self, LPFDU765A fdc) {
	_ses_start_common_session(self, fdc);
	self->name			= "INVALID";
	self->read_data		= _ses_read_result_data;
	fdc->current		= self;

	// 1) enter == RESULT PHASE ==
	self->fdc->result.push0(&self->fdc->result, 0x80);
	fdc->enter_result_phase(fdc);
}
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif

#ifdef __cplusplus
}	// extern "C"
#endif
